/* Main.c 
*/	

#include <windows.h>
#include <mmsystem.h>
#include <winuser.h>
#include <gl\gl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "project.h"
#include "resource.h"
#include "terrain.h"

// disable data conversion warnings for gl
#pragma warning(disable : 4224)
#pragma warning(disable : 4244)
#pragma warning(disable : 4305)

float totalTime=0;
int numFrame = 0;

global_t globals;
viewport_t viewport;

terrain_t *ter = NULL;


BOOL done;

BOOL drawScene( double dTime ) {
	BOOL ret = TRUE;

	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);

	glPushMatrix();

		glTranslatef( 0.0, -20.0+viewport.rot[0]/90.0*20.0, -80.0 );

		glRotatef(viewport.rot[0],1.0f,0.0f,0.0f);
		glRotatef(viewport.rot[1],0.0f,1.0f,0.0f);
		if( viewport.zoom != 1.0 );
			glScalef( viewport.zoom, viewport.zoom, viewport.zoom );

		// call render function here
		if( ter )
		{
			if( globals.rendermode == IDC_WIRE )
				drawTerrainPoint(ter);
			else
				drawTerrainPoly( ter );
		}

	glPopMatrix();

	glFlush();
	SwapBuffers( globals.hglDC );

	return ret;
}

void GLinit(void) {

	int i;
	float l_ambient[4]  = {0.2, 0.2, 0.2, 0.2};
	float l_diffuse[4]  = {0.4, 0.4, 0.4, 0.4};
	float l_specular[4] = {0.0, 0.0, 0.0, 0.0};

	glViewport(0, 0, globals.glwidth, globals.glheight);

	glLineWidth( 3 );

    /* set viewing projection */
    glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-4.0f, 4.0f, -2.704225f, 2.704225f, 4.0f, 4096.0f);  //GLQuake Mode

    /* position viewer */
	glMatrixMode(GL_MODELVIEW);

	glTranslatef( 0.0, -4.0, -12.0f );

	glLightfv( GL_LIGHT0, GL_AMBIENT, l_ambient );
	glLightfv( GL_LIGHT0, GL_DIFFUSE, l_diffuse );
	glLightfv( GL_LIGHT0, GL_SPECULAR, l_specular );
	glEnable( GL_LIGHT0 );

	viewport.lightdir[0] = 0.0;viewport.lightdir[1] = 0.0;viewport.lightdir[2] = 1.0;

	glEnable( GL_COLOR_MATERIAL );

	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);

	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);

	glShadeModel(GL_SMOOTH);

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glEnable(GL_TEXTURE_2D);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);

	for(i=0;i<3;i++)
	{
		viewport.pos[i] = 0;
		viewport.rot[i] = 0;
	}
	viewport.rot[0] = 10;
	viewport.pos[1] = 20.0;

	srand((unsigned)time( NULL ) );
	ter = NULL;

}

void idleFunc(float deltaTime)
{
	
}

HGLRC setupGL( void ) {
    static PIXELFORMATDESCRIPTOR pfd = {
        sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
        1,                              // version number
        PFD_DRAW_TO_WINDOW              // support window
        |  PFD_SUPPORT_OPENGL           // support OpenGL
        |  PFD_DOUBLEBUFFER ,           // double buffered
        PFD_TYPE_RGBA,                  // RGBA type
        16,                             // 16-bit color depth
        0, 0, 0, 0, 0, 0,               // color bits ignored
        0,                              // no alpha buffer
        0,                              // shift bit ignored
        0,                              // no accumulation buffer
        0, 0, 0, 0,                     // accum bits ignored
        16,                             // 16-bit z-buffer      
        0,                              // no stencil buffer
        0,                              // no auxiliary buffer
        PFD_MAIN_PLANE,                 // main layer
        0,                              // reserved
        0, 0, 0                         // layer masks ignored
    };
    int  pixelFormat;
    HGLRC rv = 0;

    pixelFormat = ChoosePixelFormat(globals.hglDC, &pfd);
    if ( pixelFormat ) {
        if ( SetPixelFormat(globals.hglDC, pixelFormat, &pfd) ) {
            rv = wglCreateContext( globals.hglDC );
            if ( rv ) {
                if ( !wglMakeCurrent( globals.hglDC, rv ) ) {
                    wglDeleteContext( rv );
                    rv = 0;
                }
            }
        }
    }
    return rv;
}

LONG WINAPI MainWndProc ( HWND    hWnd,
                          UINT    uMsg,
                          WPARAM  wParam,
                          LPARAM  lParam ) {
    LONG    lRet = 1;
	static int lastX = 0;
	static int lastY = 0;

    switch ( uMsg ) {
    case WM_CLOSE:
        DestroyWindow( hWnd );
        break;
    case WM_DESTROY:
        PostQuitMessage( 0 );
		break;
    case WM_QUIT:
        done = TRUE;
        break;
	case WM_SIZE:
		globals.winwidth = LOWORD( lParam );
		globals.winheight = HIWORD( lParam );
		globals.glwidth = globals.winwidth - 153;
		globals.glheight = globals.winheight-2;
		MoveWindow( globals.hGLWnd, 1,1, globals.glwidth, globals.glheight, TRUE );
		MoveWindow( globals.hMenuWnd, globals.winwidth - 152, 1, 150, globals.winheight, TRUE );
		SetFocus( globals.hWnd );
		glViewport(0, 0, globals.glwidth, globals.glheight);
		drawScene(0);
		return 0;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			BeginPaint(hWnd, &ps);
			EndPaint(hWnd, &ps);
		}
		break;
	case WM_LBUTTONDOWN:
		lastX = LOWORD(lParam);
		lastY = HIWORD(lParam);
		break;
	case WM_RBUTTONDOWN:
		lastX = LOWORD(lParam);
		lastY = HIWORD(lParam);
		break;
	case WM_MOUSEMOVE:
		if( globals.mode == MODE_BENCHSING || globals.mode == MODE_BENCHALL )
			break;
		if( wParam & MK_LBUTTON )
		{
			// rotation
			viewport.rot[1] += LOWORD(lParam)-lastX;
			if( viewport.rot[1] < 0 )
				viewport.rot[1] += 360;
			else if( viewport.rot[1] >=360 )
				viewport.rot[1] -= 360;
			// zoom
			viewport.zoom -= (float)(HIWORD(lParam)-lastY) / 50;;
			if( viewport.zoom < 1 )
				viewport.zoom = 1;
			else if( viewport.zoom >= 10.0 )
				viewport.zoom = 10.0;
			lastX = LOWORD(lParam);
			lastY = HIWORD(lParam);
			drawScene(0);
		}
		if( wParam & MK_RBUTTON )
		{
			// rotation
			viewport.rot[0] += HIWORD(lParam)-lastY;
			if( viewport.rot[0] < 0 )
				viewport.rot[0] = 0;
			else if( viewport.rot[0] >=90 )
				viewport.rot[0] = 90;
			lastX = LOWORD(lParam);
			lastY = HIWORD(lParam);
			drawScene(0);
		}
		break;
	default:
        lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }
    return lRet;
}

void benchmark( global_t *config, viewport_t *view )
{
	int i;
	for( i=0;i<3;i++ )
	{
		view->rot[i] = 0.0;
		view->pos[i] = 0.0;
	}
	EnableWindow( config->hMenuWnd, FALSE );
	config->mode = MODE_BENCHSING;
	config->benchLast = 0;
}

BOOL CALLBACK ConfigDlg( HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam )
{
	char buf[30];
    switch ( iMsg ) {
	case WM_INITDIALOG:
		CheckDlgButton( hDlg, IDC_RIVER, globals.briver );
		CheckDlgButton( hDlg, IDC_TWOD, globals.twod );
		CheckRadioButton (hDlg, IDC_SOLID, IDC_WIRE, globals.rendermode);
		sprintf( buf, "%d", globals.rseed );
		SetDlgItemText( hDlg, IDC_RSEED,  buf);
		sprintf( buf, "%f", globals.terrainscale );
		SetDlgItemText( hDlg, IDC_YSCALE,  buf);
		sprintf( buf, "%f", globals.fscale );
		SetDlgItemText( hDlg, IDC_FSCALE,  buf);
		sprintf( buf, "%d", globals.initfreq );
		SetDlgItemText( hDlg, IDC_IFREQ,  buf);
		sprintf( buf, "%f", globals.fyscale );
		SetDlgItemText( hDlg, IDC_FYSCALE,  buf);
		sprintf( buf, "%d", globals.numpass );
		SetDlgItemText( hDlg, IDC_NPASS,  buf);
		break;
	case WM_COMMAND:
		switch( LOWORD (wParam) )
		{
		case IDC_GENERATE:
			// generate a new model
			if( ter != NULL )
				freeTerrain( ter );
			ter = calloc( sizeof(terrain_t), 1);
			initTerrain( ter, 100, 100,
				150.0, 150.0,
				globals.terrainscale );
			createTerrain( ter,
				globals.rseed,
				globals.fyscale, globals.fscale, globals.initfreq, globals.numpass,
				globals.briver==BST_CHECKED ? 1: 0,
				globals.twod==BST_CHECKED ? 1: 0 );
			drawScene(0);
			drawScene(0);
			return TRUE;
		case IDC_REDRAW:
			drawScene(0);
			return TRUE;
		case IDC_BENCH:
			// call a benchmark function
			benchmark( &globals, &viewport );
			return TRUE;
		// other functionality code goes here
		case IDC_RSEED:
			switch( HIWORD(wParam)) {
			case EN_UPDATE:
			case EN_CHANGE:
				GetDlgItemText( hDlg, IDC_RSEED, buf, 20 );
				globals.rseed = atoi( buf );
				break;
			}
			return TRUE;
		case IDC_YSCALE:
			switch( HIWORD(wParam)) {
			case EN_UPDATE:
			case EN_CHANGE:
				GetDlgItemText( hDlg, IDC_YSCALE, buf, 20 );
				globals.terrainscale = atof( buf );
				break;
			}
			return TRUE;
		case IDC_FSCALE:
			switch( HIWORD(wParam)) {
			case EN_UPDATE:
			case EN_CHANGE:
				GetDlgItemText( hDlg, IDC_FSCALE, buf, 20 );
				globals.fscale = atoi( buf );
				break;
			}
			return TRUE;
		case IDC_IFREQ:
			switch( HIWORD(wParam)) {
			case EN_UPDATE:
			case EN_CHANGE:
				GetDlgItemText( hDlg, IDC_IFREQ, buf, 20 );
				globals.initfreq = atoi( buf );
				break;
			}
			return TRUE;
		case IDC_FYSCALE:
			switch( HIWORD(wParam)) {
			case EN_UPDATE:
			case EN_CHANGE:
				GetDlgItemText( hDlg, IDC_FYSCALE, buf, 20 );
				globals.fyscale = atof( buf );
				break;
			}
			return TRUE;
		case IDC_NPASS:
			switch( HIWORD(wParam)) {
			case EN_UPDATE:
			case EN_CHANGE:
				GetDlgItemText( hDlg, IDC_NPASS, buf, 20 );
				globals.numpass = atoi( buf );
				break;
			}
			return TRUE;
		case IDC_RIVER:
			globals.briver = (globals.briver==BST_CHECKED) ? BST_UNCHECKED : BST_CHECKED;
			CheckDlgButton (hDlg, IDC_RIVER, globals.briver);
			return TRUE;
		case IDC_TWOD:
			globals.twod = (globals.twod==BST_CHECKED) ? BST_UNCHECKED : BST_CHECKED;
			CheckDlgButton (hDlg, IDC_TWOD, globals.twod);
			drawScene(0);
			return TRUE;
		case IDC_SOLID:
		case IDC_WIRE:
			globals.rendermode = LOWORD(wParam);
			CheckRadioButton (hDlg, IDC_SOLID, IDC_WIRE, globals.rendermode);
			drawScene(0);
			return TRUE;
		}
		break;
	case WM_PAINT:
		break;
	}

	return FALSE;
}

int WINAPI WinMain ( HINSTANCE hInstance, 
                     HINSTANCE hPrevInstance, 
                     LPSTR lpCmdLine, 
                     int nCmdShow ) {
	WNDCLASS   wc;
    MSG msg;

	float thisTimeGetTime, lastTimeGetTime, deltaTime, nextDraw, nextInput, dTimeFrame;
	// FPS stuff

    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC)MainWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon (hInstance, "Spectral Synthesis Terrain");
    wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = "Spectral Synthesis Terrain";
    wc.lpszClassName = "Spectral Synthesis Terrain";

	// initialize dialog settings
	globals.briver = BST_UNCHECKED;
	globals.terrainscale = 8.0;
	globals.rseed = 0;
	globals.fscale = 6;
	globals.initfreq = 4;
	globals.fyscale = 0.4;
	globals.numpass = 5;
	globals.rendermode = IDC_SOLID;
	globals.twod = BST_UNCHECKED;

    if ( !RegisterClass( &wc ) ) {
        MessageBox( 0, "Couldn't Register Window Class", "Spectral Synthesis Terrain", MB_ICONERROR );
    }

    globals.hWnd = CreateWindow ("Spectral Synthesis Terrain",
                         "Spectral Synthesis Terrain",
						 WS_OVERLAPPEDWINDOW,
						 CW_USEDEFAULT, CW_USEDEFAULT,
                         CW_USEDEFAULT, CW_USEDEFAULT,
						 NULL, NULL,
                         hInstance,
                         NULL);
    if (!globals.hWnd) {
        MessageBox( 0, "Couldn't Create Window", "Spectral Synthesis Terrain", MB_ICONERROR);
    }

	globals.hGLWnd = CreateWindow("static",
                         NULL,
						 WS_CHILD | WS_VISIBLE | WS_BORDER,
                         0, 0,
						 0, 0,
                         globals.hWnd,
                         (HMENU) 10,
                         hInstance,
                         NULL);

	globals.hMenuWnd = CreateDialog( hInstance, "ConfigDlg", globals.hWnd, ConfigDlg );

	globals.hglDC  = GetDC( globals.hGLWnd );
	globals.glRC = setupGL();
	if ( !globals.glRC ) {
		MessageBox( 0, "Failed to Create OpenGL Rendering Context.",
			"GLConsole Test", MB_ICONERROR );
		DestroyWindow( globals.hWnd );
	}

    ShowWindow( globals.hWnd, SW_SHOWDEFAULT );
	ShowWindow( globals.hGLWnd, SW_SHOWDEFAULT );
	ShowWindow( globals.hMenuWnd, SW_SHOWDEFAULT );
	UpdateWindow( globals.hWnd);
	UpdateWindow( globals.hGLWnd);
	UpdateWindow( globals.hMenuWnd);
	SetFocus ( globals.hWnd);

	GLinit();

	lastTimeGetTime = (float) timeGetTime();
	deltaTime = nextInput = nextDraw = dTimeFrame = 0;

	timeBeginPeriod(1);
	srand(timeGetTime());

	drawScene(0);

    while( !done ) {
        while( PeekMessage( &msg, 0, 0, 0, PM_REMOVE ) ) {
			TranslateMessage( &msg );
			DispatchMessage( &msg );
			if(msg.message == WM_QUIT)
				done = 1;
		}
		thisTimeGetTime = (float) timeGetTime();

		if(thisTimeGetTime - lastTimeGetTime < 5)
			continue;

		deltaTime = (float)(thisTimeGetTime - lastTimeGetTime);
		dTimeFrame += deltaTime;
		lastTimeGetTime = thisTimeGetTime;
		totalTime += deltaTime;

		if( globals.mode == MODE_INTERACT )
		{
			if(dTimeFrame > 25)
			{
				if( !drawScene(totalTime) )
					done = TRUE;
				numFrame++;
				dTimeFrame = 0;
			}
		}
		else if( globals.mode == MODE_BENCHSING )
		{
			globals.benchLast += deltaTime;
			viewport.rot[1] += 5;
			drawScene(0);
			if( viewport.rot[1] == 360 )
			{
				char buf[20];
				globals.mode = MODE_NORMAL;
				viewport.rot[1] = 0;
				globals.benchLast = (float) 72*1000 / globals.benchLast;
				sprintf( buf, "Fps: %f", globals.benchLast );
				SetDlgItemText( globals.hMenuWnd, IDC_BENCH_RES, buf);
				EnableWindow( globals.hMenuWnd, TRUE );
			}
		}
	}
	timeEndPeriod(1);

	if( globals.hglDC )
		ReleaseDC( globals.hWnd, globals.hglDC );
	wglDeleteContext( globals.glRC );

	ShowCursor (TRUE);

	return msg.wParam;
}